-- | "EnrichedProlongation" contains code to attach chords to prolongation trees


module EnrichedProlongation where
import System.Random
import Chords
import Prolongation



data NoTraverse a = NoTraverse {nt_data:: a}
      
instance Show (NoTraverse a) where
    show x = ""

data TwoWayEnrichedProlongationTree a = E_Tree {
     e_vertex :: ProlongationVertex,
     e_branch :: Branch,
     e_child_l :: TwoWayEnrichedProlongationTree a,
     e_child_r :: TwoWayEnrichedProlongationTree a,
     e_ntfather :: NoTraverse (TwoWayEnrichedProlongationTree a), 
     e_fatherbranch :: Side
    } 
    | TreeChord a
    | TreeBase  
    deriving (Show) 

e_father :: TwoWayEnrichedProlongationTree a -> TwoWayEnrichedProlongationTree a
e_father t = nt_data $ e_ntfather t

data Side = S_Left | S_Right | S_Single
          deriving (Show)

arch :: Side -> TwoWayEnrichedProlongationTree a -> [TwoWayEnrichedProlongationTree a]
arch = archup


-- |might be null, but if not then should take last element?  could just pass a maybe...wil see
--can assume branch direction is the direction you want to go in
--need everthing for chord checking though
archup :: Side  -> TwoWayEnrichedProlongationTree a -> [TwoWayEnrichedProlongationTree a]
archup S_Right t@E_Tree {
               e_fatherbranch=S_Left
             } = t : (archtop S_Right (e_father t))
archup S_Right t@E_Tree {
               e_fatherbranch=S_Right
             } = t : (archup S_Right (e_father t))
archup S_Left t@E_Tree {
               e_fatherbranch=S_Right
             } = t : (archtop S_Left (e_father t))
archup S_Left t@E_Tree {
               e_fatherbranch=S_Left
             } = t : (archup S_Left (e_father t))
archup _ (TreeChord x) = [TreeChord x]
archup _ _ = []


archtop :: Side  -> TwoWayEnrichedProlongationTree a -> [TwoWayEnrichedProlongationTree a]
archtop S_Right t = archdown S_Left (e_child_r t)
archtop S_Left t = archdown S_Right (e_child_l t)
-- ^no chance of hitting a treechord here...

archdown :: Side  -> TwoWayEnrichedProlongationTree a -> [TwoWayEnrichedProlongationTree a]
archdown S_Right t@E_Tree{} = t : (archdown S_Right (e_child_r t))
archdown S_Left t@E_Tree{} = t : (archdown S_Left (e_child_l t))
archdown _ (TreeChord x) = [TreeChord x]


pTree_to_eTree ::  RandomGen a => ProlongationTree -> a-> TwoWayEnrichedProlongationTree C_ChordRep
pTree_to_eTree  p_tree g = pTree_to_eTreeTraverse p_tree TreeBase S_Single g

pTree_to_eTreeTraverse ::RandomGen a=> ProlongationTree -> TwoWayEnrichedProlongationTree C_ChordRep -> Side -> a->TwoWayEnrichedProlongationTree C_ChordRep
pTree_to_eTreeTraverse (p_tree@P_Tree{}) father side_of_father g = e_tree 
    where e_tree = E_Tree {
                e_vertex = vertex p_tree,
                e_branch = branch p_tree,
                e_child_l = pTree_to_eTreeTraverse (child_l p_tree) e_tree S_Left g1,
                e_child_r = pTree_to_eTreeTraverse (child_r p_tree) e_tree S_Right g2,
                e_ntfather = NoTraverse father,
                e_fatherbranch = side_of_father
              }
          (g1,g2) = split g
--insert chord here
pTree_to_eTreeTraverse NullTree father@E_Tree{} side_of_father g = TreeChord $ c
    where p=realParent side_of_father father
          c = case p of
                TreeBase -> randomChord g -- should pick fully random chord here
                t@E_Tree{e_vertex=Prolong_Strong} -> chordAt t
                t@E_Tree{e_vertex=Prolong_Weak} -> randomInversion (chordAt t) g
                t@E_Tree{e_vertex=Prog} -> randomChord g -- very unlikely to be identical or an inversion of the others...


-- |retrns a list of chords
treechords :: TwoWayEnrichedProlongationTree a -> [a]
treechords (TreeChord x) = [x]
treechords E_Tree{ e_child_l = l, e_child_r = r} = (treechords l) ++ (treechords r)



-- |gets the dominating chord at the current vertex
chordAt :: TwoWayEnrichedProlongationTree a -> a
chordAt (TreeChord a) = a
chordAt (E_Tree {
              e_branch=b,
              e_child_l=l,
              e_child_r=r
            }) = if b==B_Left 
                then chordAt r
                else chordAt l


scoreTree :: TwoWayEnrichedProlongationTree C_ChordRep -> Int
-- ^say want to minimize overall dissonance
--treating all vertices as progs for the moment
--the higher the score, the better the tree
--add relatedness checking for arch as well...
scoreTree (TreeChord _ ) = 0
scoreTree t@E_Tree{
              e_child_l=l,
              e_child_r=r,
              e_branch=B_Right} = subordinate (chordAt r) (chordAt l) + scoreTree r + scoreTree l
scoreTree t@E_Tree{
              e_child_l=l,
              e_child_r=r,
              e_branch=B_Left} = subordinate (chordAt l) (chordAt r) + scoreTree r + scoreTree l

-- |the higher, the worse
globRelatednessConstraints :: TwoWayEnrichedProlongationTree C_ChordRep -> Int
globRelatednessConstraints (TreeChord _ ) = 0
globRelatednessConstraints t@E_Tree{
              e_child_r=r,
              e_child_l=l,
              e_branch=B_Right} = length conflicts + globRelatednessConstraints r + globRelatednessConstraints l 
    where
      curchord = chordAt (r)
      carch = map chordAt (arch S_Right t)
      cah = head carch
      cat = tail carch
      closeness = relatedness curchord cah
      conflicts = filter (closeness<=) (map (relatedness curchord) carch)
globRelatednessConstraints t@E_Tree{
              e_child_r=r,
              e_child_l=l,
              e_branch=B_Left} = length conflicts + globRelatednessConstraints r + globRelatednessConstraints l
    where
      curchord = chordAt (l)
      carch = map chordAt (arch S_Left t)
      cah = head carch
      cat = tail carch
      closeness = relatedness curchord cah
      conflicts = filter (closeness<=) (map (relatedness curchord) carch)


sideToBranch :: Side -> Branch
sideToBranch S_Left = B_Left
sideToBranch S_Right = B_Right


branchToSide :: Branch -> Side
branchToSide B_Left = S_Left
branchToSide B_Right = S_Right

realParent :: Side -> TwoWayEnrichedProlongationTree a -> TwoWayEnrichedProlongationTree a
realParent s t@E_Tree {} = if (sideToBranch s) == e_branch t 
                           then t
                           else realParent (e_fatherbranch t) (e_father t)
realParent _ TreeBase = TreeBase

--don't apply this to treechords!
--testVertex :: branch -> TwoWayEnrichedProlongationTree C_ChordRep -> Bool
--tesetVertex t@E_Tree{} = True
--    where
--      pc = chordAt e_parent t
--      acs = 

