module Main where
import Codec.Midi
import Data.List
import System.Random
import Prolongation
import EnrichedProlongation
import Chords

type Note = Int
type Melody = [Note]

type Chord = [Note]
type Progression = [Chord]

type MidiEvent = (Ticks, Message)

midiSkeleton :: Track Ticks -> Midi
midiSkeleton mel =  Midi {
         fileType = MultiTrack, 
         timeDiv = TicksPerBeat 200, 
         tracks = [
          [
           (0,ChannelPrefix 0),
           (0,TrackName " Grand Piano  "),
           (0,InstrumentName "GM Device  1"),
           (0,TimeSignature 4 2 24 8),
           (0,KeySignature 0 0)
          ]
          ++
          mel
          ++
          [
           (0,TrackEnd)
          ]
         ]
       }  


keydown :: Note -> MidiEvent
keydown k =  (0,NoteOn {channel = 0, key = k, velocity = 80})

firstkeyup :: Note -> MidiEvent
firstkeyup k =  (480,NoteOn {channel = 0, key = k, velocity = 0})

subsequentkeyups :: Note -> MidiEvent
subsequentkeyups k =  (0,NoteOn {channel = 0, key = k, velocity = 0})
 
playchord :: Chord -> Track Ticks
playchord c = (map keydown c) ++ [firstkeyup (head c) ] ++ (map subsequentkeyups (tail c))

createMidi :: FilePath -> Progression -> IO()
createMidi f notes = exportFile  f $ midiSkeleton $ concat $ map  playchord notes


createMidis :: FilePath -> [Progression] -> IO()
createMidis f progs = sequence_ commands
    where
      indexed = zip [1,2..] progs
      commands = map (\(x,y)-> createMidi (f ++ (show x) ++ ".mid") y) indexed
      

mapmap :: (a -> b) -> [[a]] -> [[b]]
mapmap _ []     = []
mapmap f (x:xs) = map f x : mapmap f xs

mapmapmap :: (a -> b) -> [[[a]]] -> [[[b]]]
mapmapmap _ []     = []
mapmapmap f (x:xs) = mapmap f x : mapmapmap f xs

main = do 
  g<- newStdGen
  let (ga,gb) = split g
  let gas = tail $ iterate shuffle ga
  let gbs = tail $ iterate shuffle gb
  let treedata = map  (swap pt_norm1rand 5) gbs 
  let collection = take 200 $ zip treedata gas
  let trees = map (uncurry pTree_to_eTree) collection
--  print $ map scoreTree trees
--  print $ map globRelatednessConstraints trees
  let st =  map scoreTree trees
  let gr = map globRelatednessConstraints trees
--  print st
--  print gr
  let m = maximum $ st
  let n = minimum $ gr
  print $ elemIndices m $  st
  print $ elemIndices n $ gr
  createMidis "output/mlah" (mapmapmap (50+) $ mapmap stripChordRep $ map treechords trees)
              
  
-- createMidi "blah.mid" [[41,45],[42,45],[43,45],[44,46],[45,46]]